/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.7 $";
#endif

/*
 * hwgraph(3D) related functions.
 *
 * Currently (IRIX 6.5) /hw provides insufficient info to replace
 * getinvent(3) so we use it for specific device information probes only.
 *
 * A trace of hinv(1m) reveals it's usage of both getinvent() and /hw.
 *
 * See sys/iograph.h for INFO_LBL_* values.  The interesting ones look like:
 * 	INFO_LBL_NIC		Part#, Serial#, Revision, etc (TEXT)
 *	INFO_LBL_DETAIL_INVENT	Inventory binary data in unknown format.
 *	_DEVNAME_ATTR		Canonical device name in /hw
 */

#include "defs.h"
#include "myirixdeclare.h"
#include <ftw.h>
#include <sys/stat.h>
#include <sys/invent.h>
#include <sys/iograph.h>
#include <sys/attributes.h>

#define _PATH_HW_MODULE	"/hw/module"
#define _PATH_HW_NODE	"/hw/node"
#define _PATH_HW_CPUNUM	"/hw/cpunum"

static DevInfo_t	      **TreePtrPtr;
static char		      **SearchNames;
static char		       *HwDir;
static int			FindInvClass;

/*
 * Inventory types
 */
union {
    invent_generic_t		Generic;
    invent_cpuinfo_t		Cpu;
    invent_meminfo_t		Mem;
    invent_meminfo_t		Misc;
    invent_routerinfo_t		Router;
} Inv;

/*
 * Get the master's name from the child Path
 */
static char *HwgMasterName(char *Path)
{
    char		       *cp;
    char		       *Master;

    if (!Path)
	return (char *) NULL;

    Master = strdup(Path);
    if (cp = strrchr(Master, '/'))
	*cp = CNULL;

    SImsg(SIM_DBG, "HwgMasterName: Child=<%s> Master=<%s>", Path, Master);

    return Master;
}

/*
 * Get info about a CPU
 */
static DevInfo_t *HwgCPU(DevName, InvCpuInfo)
     char		       *DevName;
     invent_cpuinfo_t	       *InvCpuInfo;
{
    DevInfo_t		       *DevInfo;
    CpuInfo_t		       *CpuInfo;
    Define_t		       *Def;
    char		       *Master;
    char		       *cp;

    if (InvCpuInfo->ic_gen.ig_invclass != INV_PROCESSOR)
	return (DevInfo_t *) NULL;

    /*
     * Make sure we don't count the same CPU twice by only proceeding
     * if our master's name is "cpu"
     */
    if (Master = HwgMasterName(DevName))
	if (cp = strrchr(Master, '/'))
	    if (!EQ(cp+1, "cpu"))
		return (DevInfo_t *) NULL;

    SImsg(SIM_DBG, 
	  "HwgCPU: <%s> Flavor=%d/0x%X fq=%d sdsize=%d sdfreq=%d cpuid=%d slice=%d",
	  DevName, 
	  InvCpuInfo->ic_cpu_info.cpuflavor,
	  InvCpuInfo->ic_cpu_info.cpuflavor,
	  InvCpuInfo->ic_cpu_info.cpufq,
	  InvCpuInfo->ic_cpu_info.sdsize,
	  InvCpuInfo->ic_cpu_info.sdfreq,
	  InvCpuInfo->ic_cpuid,
	  InvCpuInfo->ic_slice
	  );

    CpuInfo = CpuInfoNew(NULL);
    CpuInfo->Speed = (Large_t) InvCpuInfo->ic_cpu_info.cpufq;
    CpuInfo->L2Cache = CacheInfoNew(NULL);
    CpuInfo->L2Cache->Size = (Large_t) InvCpuInfo->ic_cpu_info.sdsize * MBYTES;

    if (Def = DefGet("CPUflavor", NULL, 
		     InvCpuInfo->ic_cpu_info.cpuflavor, 0)) {
	CpuInfo->Man = Def->ValStr1;
	CpuInfo->Model = Def->ValStr2;
    } else {
	SImsg(SIM_DBG, 
	      "HwgCPU: <%s> Flavor=0x%X NOT DEFINED in 'CpuFlavor' in *.cf",
	      DevName, 
	      InvCpuInfo->ic_cpu_info.cpuflavor);
    }

    DevInfo = DevInfoNew(NULL);
    DevInfo->Name = strdup(DevName);
    DevInfo->Type = DT_CPU;
    DevInfo->DevSpec = (void *) CpuInfo;
    CpuInfoFill(DevInfo);

    return DevInfo;
}

/*
 * Get info about a Memory
 */
static DevInfo_t *HwgMemory(char *DevName, invent_meminfo_t *InvMemInfo)
{
    DevInfo_t		       *DevInfo;
    DevInfo_t		       *New;
    DevInfo_t		       *List = NULL;
    DevInfo_t		       *Last = NULL;
    static char			Buff[1024];
    invent_membnkinfo_t	       *MemBank;
    register int		i;
    char		       *cp;

    if (InvMemInfo->im_gen.ig_invclass != INV_MEMORY)
	return (DevInfo_t *) NULL;

    SImsg(SIM_DBG, 
	  "HwgMemory: <%s> Size=%d Banks=%d",
	  DevName, 
	  InvMemInfo->im_size,
	  InvMemInfo->im_banks
	  );

    DevInfo = DevInfoNew(NULL);
    DevInfo->Name = strdup(DevName);
    DevInfo->Type = DT_MEMORY;
    if (cp = GetSizeStr(InvMemInfo->im_size, MBYTES))
	DevInfo->ModelDesc = strdup(cp);

    /*
     * Create one device per memory bank.
     */
    for (i = 0; i < InvMemInfo->im_banks; ++i) {
	MemBank = &InvMemInfo->im_bank_info[i];

	New = DevInfoNew(NULL);
	(void) snprintf(Buff, sizeof(Buff), "%s/bank%d", DevName, i);
	New->Name = strdup(Buff);
	New->Type = DT_MEMORY;
	if (cp = GetSizeStr(MemBank->imb_size, MBYTES))
	    New->ModelDesc = strdup(cp);

	if (!Last)
	    List = Last = New;
	else {
	    Last->Next = New;
	    Last = New;
	}
    }
    DevInfo->Slaves = List;

    return DevInfo;
}

/*
 * Get attribute details and set them in DevInfo
 */
static int HwgGetDetail(char *Path, DevInfo_t *DevInfo)
{
    static char			NicBuff[1024];
    int				AttrLen;
    char		      **Argv;
    int				Argc;
    char		       *Key;
    char		       *Value;
    register int		i;

    AttrLen = sizeof(NicBuff);
    if (attr_get(Path, INFO_LBL_NIC, NicBuff, &AttrLen, 0) == 0) {
	/*
	 * Format is:
	 *  	Key1:Value1;Key2:Value2
	 * Multiple devices can exist.
	 */
	Argc = StrToArgv(&NicBuff, ";", &Argv, 0);
	for (i = 0; i < Argc; ++i) {
	    /* Get the Key/Value pair */
	    Key = Argv[i];
	    if (Value = strchr(Key, ':'))
		*Value++ = CNULL;
	    else {
		SImsg(SIM_DBG, "HwgCheck: No Value found in <%s>", Key);
		continue;
	    }
	    Value = CleanString(Value, strlen(Value), 0); /* malloc done */
	    if (!Value)
		continue;

	    /* Check the Key/Value */
	    if (EQ(Key, "Name")) {
		DevInfo->Driver = Value;
	    } else if (EQ(Key, "Part"))
		DevInfo->Part = Value;
	    else if (EQ(Key, "Serial"))
		DevInfo->Serial = Value;
	    else if (EQ(Key, "Revision"))
		DevInfo->Revision = Value;
	    else
		AddDesc(&(DevInfo->DescList), Key, Value, 0);
	}
    } else {	
	SImsg(SIM_DBG, "attr_get(%s, %s) failed: %s",
	      Path, INFO_LBL_NIC, SYSERR);
	return -1;
    }

    return 0;
}

/*
 * Check out a hwgraph node and create a DevInfo if we can.
 */
static int HwgCheck(const char *PathName, const struct stat *Stat, 
		    int Type, struct FTW *Ftw)
{
    DevInfo_t		       *DevInfo = NULL;
    int				AttrLen;
    char		       *cp;
    char		       *Path;
    char		       *DevName = NULL;
    int				DevCount = 0;

    SImsg(SIM_DBG, "HwgCheck: Check <%s>", PathName);
    Path = (char *) PathName;

    /* Shorten up the name a bit */
    if (EQN(Path, HwDir, strlen(HwDir)))
	cp = Path + 4;
    else
	cp = Path;
    DevName = cp;

    AttrLen = sizeof(Inv);
    if (attr_get(Path, INFO_LBL_DETAIL_INVENT, 
		 (char *) &Inv, &AttrLen, 0) == 0) {
	SImsg(SIM_DBG, "HwgCheck: Found Module=%d Slot=%d Class=%d Flags=0x%X",
	      Inv.Generic.ig_module, Inv.Generic.ig_slot,
	      Inv.Generic.ig_invclass, Inv.Generic.ig_flag);

	if (!FindInvClass || FindInvClass == Inv.Generic.ig_invclass) {
	    DevInfo = NULL;
	    switch (Inv.Generic.ig_invclass) {
	    case INV_PROCESSOR:
		DevInfo = HwgCPU(DevName, &Inv.Cpu);
		break;
	    case INV_MEMORY:
		DevInfo = HwgMemory(DevName, &Inv.Mem);
		break;
	    }
	}

	if (DevInfo) {
	    HwgGetDetail(Path, DevInfo);
	    AddDevice(DevInfo, TreePtrPtr, SearchNames);
	    ++DevCount;
	}
    } else {
	SImsg(SIM_DBG, "attr_get(%s, %s) failed: %s",
	      Path, INFO_LBL_DETAIL_INVENT, SYSERR);
    }

    return 0;	/* Always return 0 to insure nftw() does not abort */
}

/*
 * Find all the CPU's we can and return a list of them.
 */
extern DevInfo_t *HwgCPUfind()
{
    DevInfo_t		       *CpuList;

    TreePtrPtr = &CpuList;
    HwDir = _PATH_HW_CPUNUM;
    FindInvClass = INV_PROCESSOR;

    SImsg(SIM_DBG, "HwgCPUfind: Looking in <%s>", HwDir);

    /*
     * nftw(3) will traverse HwDir and call HwgCheck() for each
     * file/dir found.
     */
    (void) nftw(HwDir, HwgCheck, 1000, FTW_PHYS);

    if (CpuList && !CpuList->Name)
	return CpuList->Slaves;
    else
	return CpuList;
}

/*
 * Find all the Memory we can and return a list of them.
 */
extern DevInfo_t *HwgMemFind()
{
    static DevInfo_t	       *List;

    if (List)
	return List;

    TreePtrPtr = &List;
    HwDir = _PATH_HW_MODULE;
    FindInvClass = INV_MEMORY;

    SImsg(SIM_DBG, "HwgMemFind: Looking in <%s>", HwDir);

    /*
     * nftw(3) will traverse HwDir and call HwgCheck() for each
     * file/dir found.
     */
    (void) nftw(HwDir, HwgCheck, 1000, FTW_PHYS);

    if (List && !List->Name)
	return List->Slaves;
    else
	return List;
}

#if	defined(BUILD_HWG)
/*
 * Use the hwgraph (/hw) to detect devices
 */
extern int BuildDevicesHwg(DevInfo_t **MyTreePtrPtr, char **Names)
{
    struct stat			Stat;

    /* Needed by HwgCheck() */
    TreePtrPtr = MyTreePtrPtr;
    SearchNames = Names;
    FindInvClass = 0;

    SImsg(SIM_DBG, "Build Devices from hwgraph");

    /* Display type sizes to help in development */
    SImsg(SIM_DBG, "sizeof(invent_generic_t) = %d", sizeof(invent_generic_t));
    SImsg(SIM_DBG, "sizeof(invent_meminfo_t) = %d", sizeof(invent_meminfo_t));
    SImsg(SIM_DBG, "sizeof(invent_cpuinfo_t) = %d", sizeof(invent_cpuinfo_t));
    SImsg(SIM_DBG, "sizeof(invent_miscinfo_t) = %d", 
	  sizeof(invent_miscinfo_t));
    SImsg(SIM_DBG, "sizeof(invent_routerinfo_t) = %d", 
	  sizeof(invent_routerinfo_t));
    SImsg(SIM_DBG, "sizeof(inventory_t) = %d", sizeof(inventory_t));
    SImsg(SIM_DBG, "sizeof(cpu_inv_t) = %d", sizeof(cpu_inv_t));
    SImsg(SIM_DBG, "sizeof(diag_inv_t) = %d", sizeof(diag_inv_t));

    /*
     * Choose the directory to walk
     */
    if (stat(_PATH_HW_MODULE, &Stat) == 0)
	HwDir = _PATH_HW_MODULE;
    else if (stat(_PATH_HW_NODE, &Stat) == 0)
	HwDir = _PATH_HW_NODE;
    else {
	SImsg(SIM_DBG, "BuildDevicesHwg: No /hw sub-directory found.");
	return -1;
    }

    SImsg(SIM_DBG, "BuildDevicesHwg: Looking in <%s>", HwDir);

    /*
     * nftw(3) will traverse HwDir and call HwgCheck() for each
     * file/dir found.
     */
    return nftw(HwDir, HwgCheck, 1000, FTW_PHYS);
}
#endif	/* BUILD_HWG */
