/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.6 $";
#endif

/*
 * Linux DDC framebuffer and monitor functions
 *
 * The VESA BIOS Extension (VBE) is used to obtain detailed info on
 * a single installed graphics card and the monitor attached to it.
 */

#include "defs.h"
#include "edid.h"
#include "ddcprobe/vbe.h"
#include "ddcprobe/vesamode.h"
#include "ddcprobe/lrmi.h"

/*
 * Build DDC (monitor) devices
 *
 * XXX The VBE interface does not offer any means of selecting which monitor
 * device is probed.  Hence on multi-monitor systems, only the "primary"
 * console monitor is probed.
 */
extern int BuildDDC(DevInfo_t **TreePtr, char **SearchNames)
{
    DevInfo_t		       *Card = NULL;
    DevInfo_t		       *Mon = NULL;
    static Query_t		Query;
    static DevFind_t		Find;
    struct vbe_info 	       *vbe_info = NULL;
    struct vbe_edid1_info      *edid_info = NULL;
    u_int16_t 		       *mode_list = NULL;
    char		       *Vendor = NULL;
    char		       *Model = NULL;
    char		       *Revision = NULL;
    char		       *DdcSupport = NULL;
    static char			Buff[8192];
    char		       *cp;
    register int 		i;

    /*
     * Sanity checks
     */
    if (sizeof(struct vbe_info) != 512) {
	SImsg(SIM_DBG, "BuildDDC: Error: sizeof(vbe_info)=%d",
	      sizeof(vbe_info));
	return -1;
    }

    vbe_info = vbe_get_vbe_info();
    if (!vbe_info) {
	SImsg(SIM_DBG, "BuildDDC: VESA BIOS Extensions not detected.");
	return -1;
    }

    /*
     * First check for the graphics card
     */

    /*
     * Signature
     */
    snprintf(Buff, sizeof(Buff), "%c%c%c%c %d.%d",
	     vbe_info->signature[0], vbe_info->signature[1],
	     vbe_info->signature[2], vbe_info->signature[3],
	     vbe_info->version[1], vbe_info->version[0]);
    DdcSupport = strdup(Buff);
    SImsg(SIM_DBG, "BuildDDC: <%s> detected.", DdcSupport);

    if (vbe_info->oem_name.string[0]) {
	SImsg(SIM_DBG, "BuildDDC: OEM Name=<%s>", vbe_info->oem_name.string);

	if (vbe_info->version[1] >= 3) {
	    SImsg(SIM_DBG, "BuildDDC: Card Vendor=<%s> Model=<%s> Rev=<%s>",
		  vbe_info->vendor_name.string,
		  vbe_info->product_name.string,
		  vbe_info->product_revision.string);

	    if (vbe_info->vendor_name.string[0])
		Vendor = strdup(vbe_info->vendor_name.string);
	    if (vbe_info->product_name.string[0])
		Model = strdup(vbe_info->product_name.string);
	    if (vbe_info->product_revision.string[0])
		Revision = strdup(vbe_info->product_revision.string);
	}
	if (!Vendor)
	    Vendor = strdup(vbe_info->oem_name.string);
    }

    /*
     * See if we can find an existing framebuffer which matches
     */
    DevFindInit(&Find);
    Find.Expr = DFE_AND;
    Find.Flags = DFF_REGEX;
    Find.Tree = *TreePtr;
    Find.DevType = DT_FRAMEBUFFER;
    if (Vendor) {
	snprintf(Buff, sizeof(Buff), "%s", strlower(Vendor));
	if (cp = strchr(Buff, ' ')) {
	    /* Insert the regex */
	    snprintf(cp, sizeof(Buff) - (cp - &Buff[0]) - 3, ".*");
	    Find.Vendor = Buff;
	    SImsg(SIM_DBG, "BuildDDC: Looking for existing FB Vendor=<%s>",
		  Find.Vendor);
	}
    }
    if (Card = DevFind(&Find))
	SImsg(SIM_DBG,
	      "BuildDDC: Using existing FB Name=<%s> Vendor=<%s> Model=<%s>",
	      PRTS(Card->Name), PRTS(Card->Vendor), PRTS(Card->Model));
    else {
	SImsg(SIM_DBG,
	      "BuildDDC: No existing FB found for Vendor=<%s>",
	      PRTS(Vendor));
	Card = DevInfoNew(NULL);
	Card->Type = DT_FRAMEBUFFER;
    }

    /*
     * Now that we have Card setup, add what we know already
     */
    if (Vendor && !Card->Vendor)
	Card->Vendor = Vendor;
    if (Model && !Card->Model)
	Card->Model = Model;
    if (Revision && !Card->Revision)
	Card->Revision = Revision;
    AddDesc(&(Card->DescList), "Supports", DdcSupport, 0);

    /*
     * Calculate memory size of card
     */
    SImsg(SIM_DBG, "BuildDDC: Memory installed = %d * 64k blocks = %dkb",
	  vbe_info->memory_size, vbe_info->memory_size * 64);
    if (strcasestr(vbe_info->oem_name.string, "intel") &&
	strstr(vbe_info->oem_name.string, "810")) {
	SImsg(SIM_DBG,
	      "BuildDDC: Intel 810 detected. Forcing 4MB video memory.");
	if (cp = GetSizeStr((Large_t)4, MBYTES))
	    Card->Capacity = strdup(cp);
    } else {
	if (cp = GetSizeStr((Large_t)(vbe_info->memory_size * 64), KBYTES))
	    Card->Capacity = strdup(cp);
    }

    /* 
     * List supported standard modes. 
     */
    mode_list = vbe_info->mode_list.list;
    if (*mode_list != 0xffff) {
	Buff[0] = CNULL;
	for ( ; *mode_list != 0xffff; mode_list++) {
	    for (i = 0; known_vesa_modes[i].x != 0; i++) {
		if (known_vesa_modes[i].number == *mode_list) {
		    if (Buff[0])
			StrAppend(Buff, sizeof(Buff), " ");
		    StrAppend(Buff, sizeof(Buff), 
			      known_vesa_modes[i].text);
		}
	    }
	}
	if (Buff[0])
	    AddDesc(&(Card->DescList), "Supported Std Modes", strdup(Buff), 0);
    }

    /*
     * Time to add the card to the Tree if we found anything worthwhile
     */
    if (Card->Vendor || Card->Model) {
	if (Card->Unit < 0)
	    Card->Unit = UnitAssign(DT_FRAMEBUFFER, -1, NULL);
	if (!Card->Name) {
	    snprintf(Buff, sizeof(Buff), "fb%d", Card->Unit);
	    Card->Name = strdup(Buff);
	    /* 
	     * Assume that if it doesn't have a device Name, it hasn't been
	     * added to the device tree yet.
	     */
	    AddDevice(Card, TreePtr, SearchNames);
	}
    }

    /*
     * Get EDID Monitor data
     */
    if (!vbe_get_edid_supported()) {
	SImsg(SIM_DBG, "BuildDDC: EDID read not supported by video card.");
	return 0;
    }
    
    edid_info = vbe_get_edid_info();

    /*
     * Result checks
     */
    if ((edid_info == NULL) || (edid_info->version == 0)) {
	SImsg(SIM_DBG,
	      "BuildDDC: EDID read failed (No DDC-capable monitor attached?)");
	return 0;
    }
    if ((edid_info->version == 0xff && edid_info->revision == 0xff) ||
	(edid_info->version == 0 && edid_info->revision == 0)) {
	SImsg(SIM_DBG, "BuildDDC: EDID read failed, returned empty data");
	return 0;
    }

    /*
     * Set the DataLen based on version
     */    
    memset(&Query, CNULL, sizeof(Query));
    Query.Data = (void *) edid_info;
    switch (edid_info->version) {
    case 1:
	    Query.DataLen = EDID_V1_SIZE;
	    break;
    case 2:
	    Query.DataLen = EDID_V2_SIZE;
	    break;
    default:
	SImsg(SIM_DBG, "BuildDDC: Unsupported EDID Version=%d",
	      edid_info->version);
	return 0;
    }

    /*
     * Decode the EDID data.  Results are returned in Query.DevInfo.
     */
    if (EdidDecode(&Query) == 0) {
	/*
	 * Attach this monitor to it's FB card
	 */
	if (Mon = Query.DevInfo) {
	    Mon->Master = Card;
	    Mon->MasterName = Card->Name;
	    AddDevice(Mon, TreePtr, SearchNames);
	}
    }

    return 0;
}
